// EDDBEDIT.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "campaignobj.h"
#include "mapobject.h"
#include "maphexobject.h"
#include "politicalobject.h"
#include "characterobject.h"
#include "shipobject.h"
#include "auctionobject.h"
#include "newsobject.h"
#include "prelimdata.h"

struct sMiscObject {
	eMiscObjectType objtype;
	tNewsObject* news;
	tPrepObject* prep;
	tShipObject* ship;
	tCharacterObject* character;
	tAuctionObject* bid;
};
typedef tMapHexObject *MapHexPtr;


void lGetFileName(std::string& fname, char *ftypename);
void lSelectCampaignType();
bool lGetObject(ifstream& fpin, sMiscObject& obj);
void lGenerateReport( tPrelimData prelim,  tCampaignObject campaign,  tMapObject map,
				  const long numhexes, const MapHexPtr* maphexes,  tPoliticalObject tensionmatrix,
				  const long numobjects,  sMiscObject* objarray, ofstream& fpout, const bool *objdelflags, const long arrsize);
void lCleanDatabase(tPrelimData& prelim, long numhexes, long& numobjects, sMiscObject* objarray, bool *objdelflags, long arrsize);
void lEditDatabase(tPrelimData& prelim, tCampaignObject& campaign, tMapObject& map,
				  long& numhexes, MapHexPtr* maphexes, tPoliticalObject& tensionmatrix,
				  long& numobjects, sMiscObject* objarray, bool *objdelflags, long arrsize);
void lWriteDatabase(tPrelimData prelim, tCampaignObject campaign, tMapObject map,
				  long numhexes, MapHexPtr* maphexes, tPoliticalObject tensionmatrix,
				  long numobjects, sMiscObject* objarray, ofstream& fpout, bool *objdelflags, long arrsize);
void lWriteToDB(ofstream& fpout, sMiscObject obj);
void lWriteToReport(ofstream& fpout, sMiscObject obj);
sMiscObject *lLookupObject(sMiscObject *arr, long lower, long upper, long id);


int main(int argc, char* argv[])
{
	// obtain a name for the input database file (mandatory)
	std::string filename;
	//if (EDDebug) filename = "Sfc2Spd.sds"; else
	if (argc > 1) filename = argv[1];
	else lGetFileName(filename, "input db file");
	ifstream fpin;
	fpin.open(filename.c_str(), ios::binary);
	if (fpin.fail()) {
		cout << "Unable to open input file " << filename.c_str() << ",";
		cout << "   terminating program" << endl;
		return 2;
	}

	// set up variables in case a report file is to be generated
	bool genreport = true;
	ofstream fpreport;
	std::string reportname;

	// set up variables in case a modified db file is to be produced
	bool gendb = true;
	ofstream fpdbout;
	std::string dboutname;

	// ================================================
	// READ IN THE DATABASE AND BUILD THE LOCAL OBJECTS
	// ================================================

	// first the file and campaign data
	lSelectCampaignType();

	// if a report is supposed to be generated
	// obtain a name for the text file to contain the report
	if (genreport && gGetGenReport()) {
		if (EDDebug) reportname = "dbreport.txt";
		else if (argc > 2) reportname = argv[2];
		else lGetFileName(reportname, "report file");
		fpreport.open(reportname.c_str());
		if (fpreport.fail()) {
			cout << "Unable to open output file " << reportname.c_str() << "," << endl;
			cout << "   so a full report will not be generated" << endl;
			genreport = false;
		}
	}

	// if the database is supposed to be modified
	// obtain a name for the output database file
	if (gendb && (gGetCleanDB() || gGetEditDB())) {
		if (EDDebug) dboutname = "dbout.sds";
		else if (argc > 3) dboutname = argv[3];
		else lGetFileName(dboutname, "output db file");
		fpdbout.open(dboutname.c_str(), ios::binary);
		if (fpdbout.fail()) {
			cout << "Unable to open output database file " << dboutname.c_str() << "," << endl;
			cout << "   so an updated database cannot be produced" << endl;
			gendb = false;
		}
	} else gendb = false;

	// begin section with tPrelimData
	tPrelimData prelim(fpin);
	cout << "finished reading prelim data" << endl;

	// begin section with tCampaignObject
	tCampaignObject campaign(fpin);
	cout << "finished reading campaign data" << endl;

	// begin section with tMapObject

	// now the map
	tMapObject map(fpin);
	cout << "finished reading initial map data" << endl;

	long numhexes = map.mGetNumHexes();

	// begin section with MapHexPtr array
	MapHexPtr *maphexes = new MapHexPtr[numhexes];
	for (int i = 0; i < numhexes; i++) {
			maphexes[i] = new tMapHexObject(fpin);
	}
	cout << "finished reading map data" << endl;


	// begin section with tPoliticalObject
			
	// now the political matrix
	tPoliticalObject tensionmatrix(fpin);
	cout << "finished reading political data" << endl;

	// begin section with sMiscObject array
				
	// now the ships, characters, bids, and news
	long numobjects = prelim.mGetNumObjects() - (numhexes + 3);
	cout << numobjects << " objects specified" << endl;
	sMiscObject* objectarray = new sMiscObject[numobjects];
	long arrsize = numobjects;
	int j;
	cout << "Getting objects:" << endl;
	for (j = 0; j < arrsize; j++) {
		if (!lGetObject(fpin, objectarray[j])) {
			numobjects = j;
			break;
		}
	}
	cout << "finished with input file" << endl;

	// and close the input file
	fpin.close();

	// =============================================================
	// EDIT AND OR CLEAN THE DATABASE AND GENERATE A REVISED VERSION
	// =============================================================

	// have an array keep track of which objects have been deleted
	bool* deleteobject = new bool[arrsize];
	for (j = 0; j < arrsize; j++) {
		deleteobject[j] = false;
	}

	if (gendb) {

		if (gGetCleanDB()) {
			// prior to cleaning, mark all the objects as deleted,
			// then during cleaning we'll restore the ones we want
			for (j = 0; j < numobjects; j++) {
				deleteobject[j] = true;
			}

			// decide which objects to keep, and update the sizes in prelim and numobjects
			lCleanDatabase(prelim, numhexes, numobjects, objectarray, deleteobject, arrsize);

			// output the results to the db file and close it
			lWriteDatabase(prelim, campaign, map, numhexes, maphexes, tensionmatrix, numobjects, objectarray, fpdbout, deleteobject, arrsize);

			// determine how many objects were deleted
			int delcount = 0;
			for (int i = 0; i < numobjects; i++) {
				if (deleteobject[i] == true) delcount++;
			}
			cout << "DB CLEAN COMPLETE: removed " << delcount << " out of ";
			cout << numobjects << " objects from database" << endl << endl;
		}

		if (gGetEditDB()) {
			// make any further editing changes the user desires
			lEditDatabase(prelim, campaign, map, numhexes, maphexes, tensionmatrix, numobjects, objectarray, deleteobject, arrsize);

			// output the results to the db file and close it
			lWriteDatabase(prelim, campaign, map, numhexes, maphexes, tensionmatrix, numobjects, objectarray, fpdbout, deleteobject, arrsize);
		}

		// close the opened db output file
		fpdbout.close();
	}


	if (gGetGenReport() && genreport) {
		cout << "now generating report" << endl;
		if (EDDebug) {
			cout << "hit c to continue" << endl;
			char contchar;
			cin >> contchar;
		}
		lGenerateReport(prelim, campaign, map, numhexes, maphexes, tensionmatrix, numobjects, objectarray, fpreport, deleteobject, arrsize);
		fpreport.close();
	}

	delete objectarray;

	// end section with sMiscObject array
	cout << "== finished with sMiscObject array == " << endl;

	// end section with tPoliticalObject
	cout << "== finished with tPoliticalObject ==" << endl;

	delete maphexes;

	// end section with MapHexPtr array
	cout << "== finished with MapHexPtr array ==" << endl;

	// end section with tMapObject
	cout << "== finished tMapObject ==" << endl;

	// end section with tCampaignObject
	cout << "== finished tCampaignObject ==" << endl;

	// end section with tPrelimData
	cout << "== finished tPrelimData ==" << endl;

	cout << "Exiting program" << endl;
	return 0;
}

void lGetFileName(std::string& fname, char *ftypename)
{
	cout << "Please enter the name of the " << ftypename << endl;
	char filename[256];
	cin >> filename;
	fname = filename;
	char c;
	do {;
		cin.get(c);
	}while ((c != '\n') && (c != '\r'));
}


void lGenerateReport( tPrelimData prelim,  tCampaignObject campaign,  tMapObject map,
				  const long numhexes, const MapHexPtr* maphexes,  tPoliticalObject tensionmatrix,
				  const long numobjects,  sMiscObject* objarray, ofstream& fpout, const bool *objdelflags, const long arrsize)
{

	// display the file and campaign summaries
	prelim.mWriteToReport(fpout);
	campaign.mWriteToReport(fpout);

	// display the map summary
	map.mWriteToReport(fpout);
	for (long i = 0; i < numhexes; i++) {
		maphexes[i]->mWriteToReport(fpout);
	}

	// display the political matrix summary
	tensionmatrix.mWriteToReport(fpout);

	// for the remaining objects,
	//    display them if they haven't been marked for deletion
	// (and count how many did get marked for deletion)
	long delcount = 0;
	long index = 0;
	long objcount = 0;
	while ((index < arrsize) && (objcount < numobjects)) {
		if (objdelflags[index]) {
			delcount++;
		} else {
			// cout << "Writing object " << objcount << " from position " << index << endl;
			lWriteToReport(fpout, objarray[index]);
			objcount++;
		}
		index++;
	}
	if (delcount > 0) {
		fpout << endl << "TOTAL DELETED OBJECTS: " << delcount << " of " << (arrsize + numhexes + 3);
	}
	fpout << endl << endl;
}
void lEditPlayer(long numobjects, sMiscObject* objarray, long arrsize, tCampaignObject& campaign);

void lEditShipyard(long numobjects, sMiscObject* objarray, long arrsize);

void lEditMap(tMapObject& map, long numhexes, MapHexPtr* maphexes);

void lEditDatabase(tPrelimData& prelim, tCampaignObject& campaign, tMapObject& map,
				  long& numhexes, MapHexPtr* maphexes, tPoliticalObject& tensionmatrix,
				  long& numobjects, sMiscObject* objarray, bool *objdelflags, long arrsize)
{
	bool quit = false;
	cout << "=================================================" << endl;
	cout << " DISCLAIMER: there is litte / no error checking," << endl;
	cout << "   and there is no UNDO facility for the editor," << endl;
	cout << "   so MAKE BACKUPS BEFORE EDITING and BE CAREFUL!" << endl;
	cout << "=================================================" << endl;
	do {
		cout << endl;
		cout << "============== MAIN EDIT MENU ==============" << endl;
		cout << "Enter Q to quit" << endl;
		cout << "   or P to edit a player\'s statistics or ship" << endl;
		cout << "   or T to edit the political tensions" << endl;
		cout << "   or M to edit the map" << endl;
		cout << "   or C to edit other campaign data" << endl;
		cout << "   or S to edit a ship in the shipyards" << endl;
//		if (gIsEAWCampaign()) {
//			cout << "   or O to convert the EAW campaign to an OP campaign" << endl;
//		}
		cout << "choice>" << endl;
		char choice;
		cin >> choice;
		switch (choice) {
			case 'Q': 
			case 'q': quit = true;
				break;
			case 'P':
			case 'p':
				lEditPlayer(numobjects, objarray, arrsize, campaign);
				break;
			case 'T':
			case 't':
				tensionmatrix.mModifyTensions();
				break;
			case 'S':
			case 's':
				lEditShipyard(numobjects, objarray, arrsize);
				break;
			case 'M':
			case 'm':
				lEditMap(map, numhexes, maphexes);
				break;
			case 'C':
			case 'c':
				campaign.mModifyCampaign();
				break;
			case 'O':
			case 'o':
//				if (gIsEAWCampaign()) gSetOPCampaign();
				break;
			default:
				cout << "Invalid command (" << choice << ")" << endl;
				break;
		}
	} while (quit != true);
}

sMiscObject *lFindCharacter(long numobjects, sMiscObject* objarray, char *name)
// search the object array for the matching playername and return a pointer to the player
{
	if ((objarray == NULL) || (name == NULL) || (numobjects < 1)) return NULL;
	long len = strlen(name);
	for (int i = 0; i < numobjects; i++) {
		if (objarray[i].objtype != kCharObj) continue;
		if (objarray[i].character == NULL) continue;
		if (0 == strncmp(name, objarray[i].character->mGetName(), len)) 
			return (&objarray[i]);
	}
	return NULL;
}

void lEditPlayer(long numobjects, sMiscObject* objarray, long arrsize, tCampaignObject& campaign)
{
	int i;


	if ((objarray == NULL) || (numobjects < 1) || (arrsize < numobjects)) {
		cout << "UNABLE TO EDIT PLAYER DATABASE" << endl;
		return;
	}
	cout << endl << "EDITING THE PLAYER DATABASE" << endl;
	long playerid = 0;
//	cout << "If you know the object number of the player you wish to modify,";
//	cout << "enter it now, otherwise enter 0" << endl;
//	cin >> playerid;
	sMiscObject *player;
	char playername[256];
	if (playerid == 0) {
		cout << "Enter the player name for the character to be modified" << endl;
		// the first read is just to get rid of any extra garbage in the buffer,
		//     as there *will* be some left over from previous commands
		cin.getline(playername, 255);
		cin.getline(playername, 255);
		player = lFindCharacter(numobjects, objarray, playername);
	} else { 
		player = lLookupObject(objarray, 0, numobjects-1, playerid);
	}
	if ((player == NULL) || (player->objtype != kCharObj) || (player->character == NULL)) {
		cout << "UNABLE TO FIND SPECIFIED CHARACTER IN THE DATABASE" << endl;
		return;
	} 
	long prace = player->character->race;
	bool quit = false;
	do {
		quit = player->character->mModifyCaptain();
	} while (quit == false);
	bool racechange = false;
	if (player->character->race != prace) {
		racechange = true;
		prace = player->character->race;
		if (!gIsServerDB()) campaign.mSetSPRace(prace);
	}

	// ================= SHIP EDITING SECTION ========================
	char cmd;
	bool edit = true;
	do {
		cout << endl << endl;
		cout << "Do you wish to edit the player\'s ship(s)? (Y or N)" << endl;
		cin >> cmd;
		if ((cmd == 'n') || (cmd == 'N')) edit = false;
		if ((cmd != 'y') && (cmd != 'Y')) {
			cout << "Invalid choice entered: " << cmd << endl;
		}
	} while ((cmd != 'y') && (cmd != 'Y') && (cmd != 'n') && (cmd != 'N'));

	if ((edit == false) && (racechange == false)) return;

	long numships, shipids[3];
	tShipObject *ship1, *ship2, *ship3;
	sMiscObject *holder;
	player->character->mGetShipIDs(numships, shipids);
	if (numships < 1) return;
	if (numships > 3) return;

	if (edit) cout << endl << " === NOW EDITING CHARACTERS SHIP(S) ===" << endl;
	bool updateplayer = false;
	long difference;

	// get player's first ship
	if (edit) cout << endl << " === Editing ship 1 ===" << endl;
	holder = lLookupObject(objarray, 0, numobjects-1, shipids[0]);
	if (holder) ship1 = holder->ship;
	else ship1 = NULL;
	if (ship1) {
		ship1->race = prace;
		if (edit) updateplayer = ship1->mModifyShip();
	}
	
	if ((updateplayer) && (ship1)) {
		difference = ship1->typelen + ship1->namelen;
		difference = difference - player->character->currships[0].namelen;
		difference = difference - player->character->currships[0].typelen;
		player->character->size1 += difference;
		player->character->size2 += difference;

		player->character->currships[0].shipbpv = ship1->shipbpv1;
		player->character->currships[0].hullclass = ship1->hullclass;
		player->character->currships[0].namelen = ship1->namelen;
		player->character->currships[0].typelen = ship1->typelen;
		delete player->character->currships[0].shiptype;
		delete player->character->currships[0].shipname;
		player->character->currships[0].shipname = new char[ship1->namelen];
		player->character->currships[0].shiptype = new char[ship1->typelen];
		for (i = 0; i < ship1->namelen; i++) {
			player->character->currships[0].shipname[i] = ship1->namestr[i];
		}
		for (i = 0; i < ship1->namelen; i++) {
			player->character->currships[0].shiptype[i] = ship1->typestr[i];
		}
	}

	// get player's second ship
	if (numships > 1) {
		updateplayer = false;
		ship2 = NULL;
		if (edit) cout << endl << " === Editing ship 2 ===" << endl;
		holder = lLookupObject(objarray, 0, numobjects-1, shipids[1]);
		if (holder) ship2 = holder->ship;
		if (ship2) {
			ship2->race = prace;
			if (edit) updateplayer = ship2->mModifyShip();
		}
		if ((updateplayer) && (ship2)) {
			difference = ship2->typelen + ship2->namelen;
			difference = difference - player->character->currships[1].namelen;
			difference = difference - player->character->currships[1].typelen;
			player->character->size1 += difference;
			player->character->size2 += difference;

			player->character->currships[1].shipbpv = ship2->shipbpv1;
			player->character->currships[1].hullclass = ship2->hullclass;
			player->character->currships[1].namelen = ship2->namelen;
			player->character->currships[1].typelen = ship2->typelen;
			delete player->character->currships[1].shiptype;
			delete player->character->currships[1].shipname;
			player->character->currships[1].shipname = new char[ship2->namelen];
			player->character->currships[1].shiptype = new char[ship2->typelen];
			for (i = 0; i < ship2->namelen; i++) {
				player->character->currships[1].shipname[i] = ship2->namestr[i];
			}
			for (i = 0; i < ship2->namelen; i++) {
				player->character->currships[1].shiptype[i] = ship2->typestr[i];
			}
		}
	}

	// get player's third ship
	if (numships > 2) {
		updateplayer = false;
		ship3 = NULL;
		if (edit) cout << endl << " === Editing ship 3 ===" << endl;
		holder = lLookupObject(objarray, 0, numobjects-1, shipids[2]);
		if (holder) ship3 = holder->ship;
		if (ship3) {
			ship3->race = prace;
			if (edit) updateplayer = ship3->mModifyShip();
		}
		if ((updateplayer) && (ship3)) {
			difference = ship3->typelen + ship3->namelen;
			difference = difference - player->character->currships[2].namelen;
			difference = difference - player->character->currships[2].typelen;
			player->character->size1 += difference;
			player->character->size2 += difference;

			player->character->currships[2].shipbpv = ship3->shipbpv1;
			player->character->currships[2].hullclass = ship3->hullclass;
			player->character->currships[2].namelen = ship3->namelen;
			player->character->currships[2].typelen = ship3->typelen;
			delete player->character->currships[2].shiptype;
			delete player->character->currships[2].shipname;
			player->character->currships[2].shipname = new char[ship3->namelen];
			player->character->currships[2].shiptype = new char[ship3->typelen];
			for (i = 0; i < ship3->namelen; i++) {
				player->character->currships[2].shipname[i] = ship3->namestr[i];
			}
			for (i = 0; i < ship3->namelen; i++) {
				player->character->currships[2].shiptype[i] = ship3->typestr[i];
			}
		}
	}

}

void lEditShipyard(long numobjects, sMiscObject* objarray, long arrsize)
{
	// ================= SHIP EDITING SECTION ========================
	long shipid;
	cin >> shipid;

	tShipObject *ship1;
	sMiscObject *holder;

	cout << endl << " === NOW EDITING SHIPYARD VESSEL ===" << endl;
	bool unusedflag = false;

	holder = lLookupObject(objarray, 0, numobjects-1, shipid);
	if (holder) ship1 = holder->ship;
	else ship1 = NULL;
	if (!ship1) {
		cout << "ERROR: unable to find shipyard vessel with id " << shipid << endl;
		return;
	} else
		unusedflag = ship1->mModifyShip();
}

void lEditMap(tMapObject& map, long numhexes, MapHexPtr* maphexes)
{
	long r, c;
	cout << endl << "EDITING THE HEX MAP" << endl;
	long rows = map.mGetRows();
	long columns = map.mGetColumns();
	if ((maphexes == NULL)|| (numhexes < (rows*columns-1))) {
		cout << "Bad map construction, unable to edit" << endl;
	}
	char cmd;
	bool quit;
	do {
		cout << "Enter A to modify ALL hexes simultaneously" << endl;
		cout << "   or R to modify a range of hexes" << endl;
		cout << "   or I to modify hexes individually" << endl;
		cout << "   or Q to quit" << endl;
		cin >> cmd;
		if ((cmd == 'Q') || (cmd == 'q')) return;
	} while ((cmd != 'I') && (cmd != 'i') && (cmd != 'A') && (cmd != 'a'));

	if ((cmd == 'R') || (cmd == 'r')) {
		long ULrow,ULcol, LRrow, LRcol;
		// get the affected hex rectangle from the user
		cout << "This edits a rectangle of hexes (with additional restrictions if desired)" << endl;
		cout << "Enter the row then column for the upper left corner of the hexes to be modified" << endl;
		cin >> ULrow >> ULcol;
		cout << "Enter the row then column for the lower right corner of the hexes to be modified" << endl;
		cin >> LRrow >> LRcol;
		// correct the entered hexes to ensure they're in bounds
		if (ULrow < 0) ULrow = 0;
		if (ULrow >= rows) ULrow = rows - 1;
		if (ULcol < 0) ULcol = 0;
		if (ULcol >= columns) ULcol = columns - 1;
		if (LRrow < ULrow) LRrow = ULrow;
		if (LRrow >= rows) LRrow = rows - 1;
		if (LRcol < ULcol) LRcol = ULcol;
		if (LRcol >= columns) LRcol = columns - 1;

		// variables to control what aspect of the hexes are changed,
		// and to restrict the type of hexes the changes are applied to
		char adjtype = 0;
		long adjamount = 0;
		long rstrtype = 0;
		long rstramount = 0;
		cout << "Select the nature of the change you want to make:" << endl;
		cout << "Enter D to set a specific DV value" << endl;
		cout << "   or d to change DVs by a set amount" << endl;
		cout << "   or E to set a specific econ value" << endl;
		cout << "   or e to change econ by a set amount" << endl;
		cout << "   or T to set a specific terrain type" << endl;
		cout << "   or P to set a specific planet type" << endl;
		cout << "   or B to set a specific base type" << endl;
		cout << "   or R to set a specific race type (empire or cartel)" << endl;

		// apply the change to every matching hex in the block
		for (r = ULrow; r <= LRrow; r++) {
			for (c = ULcol; c <= LRcol; c++) {
				maphexes[(r * columns) + c]->mModifyHex(adjtype, adjamount, rstrtype, rstramount);
			}
		}
	} else if ((cmd == 'A') || (cmd == 'a')) {
		quit = false;
		do {
			cout << "Enter R to reset all current hex DVs to fifty percent of their maximum" << endl;
			cout << "   or D to add a fixed amount to all hexes maximum DVs" << endl;
			cout << "   or d to add a fixed amount to all hexes current DVs" << endl;
			cout << "   or E to add a fixed amount to all hexes maximum Econ" << endl;
			cout << "   or e to add a fixed amount to all hexes current Econ" << endl;
			cout << "   or enter anything else to quit" << endl;
			cin >> cmd;
			if ((cmd == 'E') || (cmd == 'e') || (cmd == 'D') || (cmd == 'd') || (cmd == 'r') || (cmd == 'R')) {
				long adjustment = 0l;
				if ((cmd != 'r') && (cmd != 'R')) {
					cout << "Enter the desired adjustment value" << endl;
					cin >> adjustment;
				}
				for (int i = 0; i < numhexes; i++) {
					switch (cmd) {

						case 'R': case 'r':
							maphexes[i]->mHalveCurrDV();
							break;

						case 'D':
							maphexes[i]->mAdjustMaxDV(adjustment);
							break;

						case 'd':
							maphexes[i]->mAdjustCurrDV(adjustment);
							break;

						case 'E':
							maphexes[i]->mAdjustMaxEcon(adjustment);
							break;

						case 'e':
							maphexes[i]->mAdjustCurrEcon(adjustment);
							break;

						default:
							cout << "Invalid selection: " << cmd << endl;
							break;
					}
				}
			} else {
				quit = true;
			}
		} while (quit == false);
	} else {
		quit = false;
		do {
			do {
				cout << "Pick a row between 0 and " << (rows-1) << endl;
				cin >> r;
				if ((r < 0) || (r >= rows)) cout << "Invalid choice" << endl;
			} while ((r < 0) || (r >= rows));
			do {
				cout << "Pick a column between 0 and " << (columns-1) << endl;
				cin >> c;
				if ((c < 0) || (c >= columns)) cout << "Invalid choice" << endl;
			} while ((c < 0) || (c >= columns));
			maphexes[(r * columns) + c]->mModifyHex();
			cout << "Enter E to edit another hex," << endl;
			cout << "   or anything else to quit" << endl;
			cin >> cmd;
			if ((cmd != 'E') && (cmd != 'e')) quit = true;
		} while (quit == false);
	}
}


long lGetObjectID(sMiscObject obj)
{
	switch (obj.objtype) {
		case kShipObj:
			if (obj.ship == NULL) return -1;
			else return obj.ship->mGetShipID();
		case kNewsObj:
			if (obj.news == NULL) return -1;
			else return obj.news->mGetNewsID();
		case kPrepObj:
			if (obj.prep == NULL) return -1;
			else return obj.prep->mGetPrepID();
		case kBidObj:
			if (obj.bid == NULL) return -1;
			else return obj.bid->mGetBidID();
		case kCharObj:
			if (obj.character == NULL) return -1;
			else return obj.character->mGetCaptainID();
		default: break;
	}
	return -1;
}

sMiscObject *lLookupObject(sMiscObject *arr, long lower, long upper, long id)
// perform binary search on the array to find and return the specified object
{
	if (arr == NULL) return NULL;
	if (lower < 0) return NULL;
	if (upper < lower) return NULL;
	long middle = (upper + lower) / 2;
	long lowerid = lGetObjectID(arr[lower]);
	long upperid = lGetObjectID(arr[upper]);
	long middleid = lGetObjectID(arr[middle]);
	if (id < lowerid) return NULL;
	if (id > upperid) return NULL;
	if (id == lowerid) return (&(arr[lower]));
	if (id == upperid) return (&(arr[upper]));
	if (id == middleid) return (&(arr[middle]));
	if (id < middleid) return lLookupObject(arr, lower+1, middle-1, id);
	else return lLookupObject(arr, middle+1, upper-1, id);
}

void lCleanDatabase(tPrelimData& prelim, long numhexes, long& numobjects, 
					sMiscObject* objarray, bool *objdelflags, long arrsize)
{
	// go through the object array
	long i, j;
	tShipObject *ship;
	tCharacterObject *captain;
	tAuctionObject *bid;
	tPrepObject *prep;
	long delcount = 0;

	// by default all items (news, bids, characters, ships) are marked for deletion
	// the following cases identify the exceptions, i.e. the objects we want to keep
	for (i = 0; i < numobjects; i++) {
		bool delobj = true;
		if (objarray[i].objtype == kShipObj) {
			// if an item is a ship
			//    if it is a base or planet (any type) mark it NOT to be deleted
			//    otherwise if it is unowned (i.e. captain id is 0) mark it NOT to be deleted
			//    otherwise look up the captain and if the captain is human mark it NOT to be deleted
			ship = objarray[i].ship;
			if (ship != NULL) {
				long hullclass = ship->mGetHullClass();
				if ((hullclass >= 13) && (hullclass <= 16)) delobj = false;
				else if (hullclass == 18) delobj = false;
				else if (ship->mGetCaptainID() == 0) delobj = false;
				else {
					sMiscObject *obj = lLookupObject(objarray, 0, numobjects-1, ship->mGetCaptainID());
					if ((obj != NULL) && (obj->objtype == kCharObj) && (obj->character != NULL)) {
						if (!obj->character->mIsCaptainAI()) delobj = false;
					}
				}
			}

		} else if (objarray[i].objtype == kBidObj) {
			// if an item is an auction item
			//    mark it NOT to be deleted
			bid = objarray[i].bid;
			if (bid != NULL) delobj = false;

		} else if (objarray[i].objtype == kPrepObj) {
			// if an item is a prepared mission item
			//    mark it NOT to be deleted
			prep = objarray[i].prep;
			if (prep != NULL) delobj = false;

		} else if (objarray[i].objtype == kCharObj) {
			// if an item is a character
			//    if it is a human mark it NOT to be deleted
			//    otherwise if it owns a base (any type) mark it NOT to be deleted
			captain = objarray[i].character;
			if (captain != NULL) {
				if (!captain->mIsCaptainAI()) delobj = false;
				else {
					for (j = 0; j < captain->mGetNumShips(); j++) {
						long hull = captain->mGetShipHullclass(j);
						if ((hull >= 13) && (hull <= 16)) delobj = false;
						else if (hull == 18) delobj = false;
					}
				}
			}
		}
		if (!delobj) {
			objdelflags[i] = false;
			delcount++;
		}
	}
	
	// now count the number of objects left and update prelim
	numobjects = 0;
	for (i = 0; i < arrsize; i++) {
		if (!objdelflags[i]) numobjects++;
	}
	prelim.mSetCount(numobjects + numhexes + 3);

	cout << "After cleaning thinks there are " << prelim.mGetNumObjects() << " objects" << endl;
	cout << "   and numobjects is " << numobjects << endl;
	cout.flush();

	// note: when writing the database (next function)
	//        it will only write the objects not marked for deletion
}

void lWriteDatabase(tPrelimData prelim, tCampaignObject campaign, tMapObject map,
				  long numhexes, MapHexPtr* maphexes, tPoliticalObject tensionmatrix,
				  long numobjects, sMiscObject* objarray, ofstream& fpout, bool *objdelflags, long arrsize)
{
	prelim.mWriteToDB(fpout);
	campaign.mWriteToDB(fpout);
//	if (gIsEAWCampaign()) {
		campaign.mWriteCampInfo();
//	}
	map.mWriteToDB(fpout);
	for (long i = 0; i < numhexes; i++) {
		maphexes[i]->mWriteToDB(fpout);
	}
	tensionmatrix.mWriteToDB(fpout);
	long index = 0;
	long objcount = 0;
	while ((index < arrsize) && (objcount < numobjects)) {
		if (!objdelflags[index]) {
			lWriteToDB(fpout, objarray[index]);
			objcount++;
		}
		index++;
	}
	cout << "Wrote " << (objcount + numhexes + 3) << " objects" << endl;
	cout.flush();
}

bool lGetObject(ifstream& fpin, sMiscObject& obj)
{
	obj.objtype = kOtherObj;
	obj.prep = NULL;
	obj.news = NULL;
	obj.character = NULL;
	obj.ship = NULL;
	obj.bid = NULL;

	// GetObject reads the dual size fields, header length,
	//    and header string for the next object,
	// it then calls GetObjectType, which identifies the type of object,
	//    and stores the result in obj.objtype
	// GetObject then calls new on the appropriate constructor,
	//    passing the file pointer and object size,
	//    storing the returned object pointer in the correct obj.field
	long size1, size2;
	size1 = gGetInt32(fpin);
	size2 = gGetInt32(fpin);
	if (size1 != size2) {
		cout << "WARNING: SIZE MISMATCH IN OBJECT ";
		cout << size1 << " vs " << size2 << endl;
		cout << "Hit any key to continue" << endl;
		char cmd;
		cin >> cmd;
		return false;
	} else if (fpin.eof()) {
		cout << "END OF FILE" << endl;
		return false;
	}
	char *header;
	long hdrlen;
	gGetString(fpin, hdrlen, header);
	obj.objtype = gGetObjectType(hdrlen, header);
	switch (obj.objtype) {
	case kNewsObj:
		if (EDDebug) cout << "Processing news item" << endl;
		obj.news = new tNewsObject(fpin, size1);
		break;
	case kPrepObj:
		if (EDDebug) cout << "Processing mission prep item" << endl;
		obj.prep = new tPrepObject(fpin, size1);
		break;
	case kShipObj:
		if (EDDebug) cout << "Processing ship" << endl;
		obj.ship = new tShipObject(fpin, size1);
		break;
	case kBidObj:
		if (EDDebug) cout << "Processing bid" << endl;
		obj.bid = new tAuctionObject(fpin, size1);
		break;
	case kCharObj:
		if (EDDebug) cout << "Processing character" << endl;
		obj.character = new tCharacterObject(fpin, size1);
		break;
	default:
		cout << "***Processing an unknown object type:" << header << endl;
		break;
	}
	return true;
}

void lWriteToDB(ofstream& fpout, sMiscObject obj)
{
	switch (obj.objtype) {
	case kPrepObj:
		if (obj.prep) obj.prep->mWriteToDB(fpout);
		break;
	case kNewsObj:
		if (obj.news) obj.news->mWriteToDB(fpout);
		break;
	case kShipObj:
		if (obj.ship) obj.ship->mWriteToDB(fpout);
		break;
	case kBidObj:
		if (obj.bid) obj.bid->mWriteToDB(fpout);
		break;
	case kCharObj:
		if (obj.character) obj.character->mWriteToDB(fpout);
		break;
	default:
		break;
	}
}

void lWriteToReport(ofstream& fpout, sMiscObject obj)
{
	switch (obj.objtype) {
	case kNewsObj:
		if (obj.news) obj.news->mWriteToReport(fpout);
		break;
	case kPrepObj:
		if (obj.prep) obj.prep->mWriteToReport(fpout);
		break;
	case kShipObj:
		if (obj.ship) obj.ship->mWriteToReport(fpout);
		break;
	case kBidObj:
		if (obj.bid) obj.bid->mWriteToReport(fpout);
		break;
	case kCharObj:
		if (obj.character) obj.character->mWriteToReport(fpout);
		break;
	default:
		break;
	}
}

void lSelectCampaignType()
{
	gInitDBType();
	char cmd;
	do {
		cout << "Is this an OP campaign file (y or n)?" << endl;
		cin >> cmd;
		if ((cmd == 'y') || (cmd == 'Y')) gSetOPCampaign();
	} while ((cmd != 'y') && (cmd != 'Y') && (cmd != 'n') && (cmd != 'N'));
	do {
		cout << "Do you wish to clean the file, edit the file or generate a report (pick one: C, E, or R)?" << endl;
		cin >> cmd;
		if ((cmd == 'C') || (cmd == 'c')) gSetRuntype('C');
		else if ((cmd == 'E') || (cmd == 'e')) gSetRuntype('E');
		else if ((cmd == 'R') || (cmd == 'r')) gSetRuntype('R');
	} while ((cmd != 'C') && (cmd != 'c') && (cmd != 'E') && (cmd != 'e')
		&& (cmd != 'R') && (cmd != 'r'));
}
